# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

import io
import os
import re
import operator
from abc import ABC
from time import sleep
from collections import OrderedDict
from typing import Any, Dict, Tuple, Union, List, Optional

from tqdm.auto import tqdm
from numpy import floor, ndarray

import torch
from torch.nn import Module
from torch.tensor import Tensor
from torch.cuda import is_available, amp
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader, SubsetRandomSampler

from comet_ml import Experiment
from comet_ml.api import APIExperiment
from comet_ml.experiment import BaseExperiment

from enchanter.utils.backend import is_scalar
from enchanter.callbacks import BaseLogger
from enchanter.callbacks import Callback
from enchanter.callbacks import CallbackManager
from enchanter.engine.saving import RunnerIO
from enchanter.engine.modules import send, get_dataset, is_tfds, tfds_to_numpy, restore_state_dict


__all__ = ["BaseRunner"]


class BaseRunner(ABC, RunnerIO):
    """
    A class for creating runners to train PyTorch models.


    Examples:

        >>> from comet_ml import Experiment
        >>> import torch
        >>> class Runner(BaseRunner):
        >>>     def __init__(self):
        >>>         super(Runner, self).__init__()
        >>>         self.model = torch.nn.Linear(10, 10)
        >>>         self.optimizer = torch.optim.Adam(self.model.parameters())
        >>>         self.experiment = Experiment()
        >>>         self.criterion = torch.nn.CrossEntropyLoss()
        >>>
        >>>     def train_step(self, batch):
        >>>         x, y = batch
        >>>         out = self.model(x)
        >>>         loss = self.criterion(out, y)
        >>>
        >>>         return {"loss": loss}

    """

    def __init__(self) -> None:
        super(BaseRunner, self).__init__()
        self.device: torch.device = torch.device("cuda" if is_available() else "cpu")
        self.model: Module = NotImplemented
        self.optimizer: Optimizer = NotImplemented
        self.scheduler: List = list()
        self.experiment: Union[BaseExperiment, BaseLogger] = NotImplemented
        self.manager = CallbackManager()
        self.callbacks: Optional[List[Callback]] = None
        self.scaler: Optional[amp.GradScaler] = None
        self.api_experiment: Optional[APIExperiment] = None

        self.global_step: int = 0
        self.non_blocking: bool = True
        self.configures: Dict[str, Any] = {"epochs": 0, "checkpoint_path": None}
        self.pbar: Union[tqdm, range] = range(self.configures["epochs"])
        self.metrics: Dict = {"train": dict(), "val": dict(), "test": dict()}
        self._loaders: Dict[str, Union[DataLoader, Any]] = dict()

    def backward(self, loss: Tensor) -> None:
        """
        calculate the gradient.
        If self.scaler is a `torch.cuda.amp.GradScaler` object, it is automatically processed by amp.

        Args:
            loss (torch.Tensor):

        Returns:
            None

        """
        if isinstance(self.scaler, amp.GradScaler):
            self.scaler.scale(loss).backward()
        else:
            loss.backward()

    def update_optimizer(self) -> None:
        """
        update optimizer

        """
        if isinstance(self.scaler, amp.GradScaler):
            self.scaler.step(self.optimizer)
            self.scaler.update()

        else:
            self.optimizer.step()

    def update_scheduler(self, epoch: int) -> None:
        """
        Method called to update the value of the scheduler.
        It is called after updating the Optimizer.

        Args:
            epoch (int): Current Epoch

        Returns:
            None

        Examples:
            >>> from torch.optim import lr_scheduler
            >>> from enchanter.tasks import ClassificationRunner
            >>> runner: BaseRunner = ClassificationRunner(
            >>>     model=...,
            >>>     optimizer=...,
            >>>     criterion=...,
            >>>     scheduler=[
            >>>         lr_scheduler.CosineAnnealingLR(...), lr_scheduler.StepLR(...)
            >>>     ]
            >>> )

        """
        for scheduler in self.scheduler:
            scheduler.step()

        self.experiment.log_metric("scheduler_lr", self.scheduler[-1].get_last_lr(), epoch=epoch)

    def train_step(self, batch: Tuple) -> Dict[str, Tensor]:
        """
        When training the neural network,
            >>> import torch.nn as nn
            >>> train_loader: DataLoader = ...
            >>> model: nn.Module = ...
            >>> criterion = ...
            >>> for x, y in train_loader:
            >>>     out = model(x)
            >>>     loss = criterion(out, y)

        this method is responsible for the above areas.

        Args:
            batch: A tuple containing data & labels get from the PyTorch DataLoader.

        Returns:
            You need to return a dictionary with the key 'loss'.

        Examples:
            >>> def train_step(self, batch):
            >>>     x, y = batch
            >>>     out = self.model(x)
            >>>     loss = nn.functional.cross_entropy(out, y)
            >>>     return {"loss": loss}

        """

    def train_end(self, outputs: List) -> Dict[str, Tensor]:
        """
        This method is executed at the end of each step of neural network training.

        Args:
            outputs:

        Returns:

        """
        return {}

    def val_step(self, batch: Tuple) -> Dict[str, Tensor]:
        """
        This method is executed at every 1 step when validating the neural net.
        See `train_step()` for help.


        Args:
            batch: A tuple containing data & labels get from the PyTorch DataLoader.

        Returns:
            You need to return a dictionary.

        """

    def val_end(self, outputs: List) -> Dict[str, Tensor]:
        """
        This method is executed at the end of each step of neural network validating.

        Args:
            outputs:

        Returns:
            You need to return a dictionary.

        """
        return {}

    def test_step(self, batch: Tuple) -> Dict[str, Tensor]:
        """
        This method is executed at every 1 step when testing the neural net.
        See `train_step()` for help.

        Args:
            batch: A tuple containing data & labels get from the PyTorch DataLoader.

        Returns:
            You need to return a dictionary.

        """

    def test_end(self, outputs: List) -> Dict[str, Tensor]:
        """
        This method is executed at the end of each step of neural network testing.

        Args:
            outputs:

        Returns:
            You need to return a dictionary.

        """
        return {}

    def train_cycle(self, epoch: int, loader: DataLoader) -> None:
        """
        This is a training loop for neural net.

        Args:
            epoch
            loader (torch.utils.data.DataLoader):

        """
        results = list()
        loader_size = len(loader)

        if is_tfds(loader):
            loader = tfds_to_numpy(loader)

        self.model.train()
        with self.experiment.train():
            for step, batch in enumerate(loader):
                self.optimizer.zero_grad()
                batch = send(batch, self.device, self.non_blocking)
                # on_step_start()
                self.global_step += 1
                outputs = self.train_step(batch)
                self.backward(outputs["loss"])
                self.update_optimizer()

                if hasattr(self.pbar, "set_postfix"):
                    per = "{:1.0%}".format(step / loader_size)
                    self.pbar.set_postfix(OrderedDict(train_batch=per), refresh=True)  # type: ignore

                outputs = {
                    key: outputs[key].detach().cpu() if isinstance(outputs[key], Tensor) else outputs[key]
                    for key in outputs.keys()
                }
                tmp = {k: outputs[k] for k in outputs.keys() if is_scalar(outputs[k])}
                self.experiment.log_metrics(tmp, step=self.global_step, epoch=epoch)
                results.append(outputs)
                self.manager.on_train_step_end(outputs)

            dic = self.train_end(results)  # pylint: disable=E1111

            if len(dic) != 0:
                self.metrics["train"].update(dic)
                self.experiment.log_metrics(dic, step=epoch, epoch=epoch)

    def val_cycle(self, epoch: int, loader: DataLoader) -> None:
        """
        This is a validating loop for neural net.

        Args:
            epoch
            loader:

        Returns:

        """
        results = list()
        loader_size = len(loader)

        if is_tfds(loader):
            loader = tfds_to_numpy(loader)

        self.model.eval()
        with self.experiment.validate(), torch.no_grad():
            for step, batch in enumerate(loader):
                batch = send(batch, self.device, self.non_blocking)
                self.global_step += 1
                # on_step_start()
                outputs = self.val_step(batch)  # pylint: disable=E1111

                if hasattr(self.pbar, "set_postfix"):
                    per = "{:1.0%}".format(step / loader_size)
                    self.pbar.set_postfix(OrderedDict(val_batch=per), refresh=True)  # type: ignore

                outputs = {
                    key: outputs[key].cpu() if isinstance(outputs[key], Tensor) else outputs[key]
                    for key in outputs.keys()
                }
                tmp = {k: outputs[k] for k in outputs.keys() if is_scalar(outputs[k])}
                self.experiment.log_metrics(tmp, step=self.global_step, epoch=epoch)
                results.append(outputs)
                self.manager.on_validation_step_end(outputs)

            dic = self.val_end(results)  # pylint: disable=E1111

            if len(dic) != 0:
                self.metrics["val"].update(dic)
                self.experiment.log_metrics(dic, step=epoch, epoch=epoch)

    def test_cycle(self, loader: DataLoader) -> None:
        """
        This is a testing loop for neural net.

        Args:
            loader:

        Returns:

        """
        results = list()
        loader_size = len(loader)

        if is_tfds(loader):
            loader = tfds_to_numpy(loader)

        self.model.eval()
        with self.experiment.test(), torch.no_grad():
            for step, batch in enumerate(loader):
                batch = send(batch, self.device, self.non_blocking)
                # on_step_start()
                outputs = self.test_step(batch)  # pylint: disable=E1111

                per = "{:1.0%}".format(step / loader_size)
                if hasattr(self.pbar, "set_postfix"):
                    self.pbar.set_postfix(OrderedDict(test_batch=per), refresh=True)  # type: ignore

                    self.pbar.update(1)  # type: ignore

                outputs = {
                    key: outputs[key].cpu() if isinstance(outputs[key], Tensor) else outputs[key]
                    for key in outputs.keys()
                }

                tmp = {k: outputs[k] for k in outputs.keys() if is_scalar(outputs[k])}
                self.experiment.log_metrics(tmp)
                results.append(outputs)
                self.manager.on_test_step_end(outputs)

            dic = self.test_end(results)  # pylint: disable=E1111

            if len(dic) != 0:
                self.metrics["test"].update(dic)
                self.experiment.log_metrics(dic)

    def train_config(
        self,
        epochs: int,
        checkpoint_path: Optional[str] = None,
        monitor: Optional[str] = None,
    ):
        """
        This method is used to specify epochs and so on when you execute using the .run() method.

        Examples:
            >>> runner: BaseRunner = ...
            >>> runner.train_config(
            >>>     epochs=10,
            >>>     checkpoint_path="/path/to/checkpoint_dir",
            >>>     monitor="validate_avg_acc >= 0.75"
            >>> )

        Args:
            epochs (int): Specify the number of training epochs.
            checkpoint_path: Specify the name of the directory where the checkpoint is stored,
                             and if monitor is not specified, store weights for all epochs.
            monitor: Save only the epoch that corresponds to the specified expression.
                     The `checkpoint_path` must be set together with it.

        Returns:
            None
        Notes:
            When you specify the monitor argument, be sure to put a space between 'keyword', 'symbol', and 'value'.

        """

        self.configures["checkpoint_path"] = checkpoint_path
        if monitor:
            try:
                _ = re.search("train|validate", monitor)[0]  # type: ignore
            except TypeError:
                raise KeyError("The argument monitor is not an expected expression. {}".format(monitor))
            else:
                self.configures["monitor"] = monitor

            if not isinstance(self.experiment, Experiment):
                raise TypeError(
                    "To use `.train_config(, monitor='....')`, you need an `Experiment` object. `experiment` is {}.\
                    ".format(
                        type(self.experiment)
                    )
                )

        if epochs > 0:
            self.configures["epochs"] = epochs
        else:
            self.configures["epochs"] = 1

        return self

    def log_hyperparams(self, dic: Dict = None, prefix: Optional[str] = None) -> None:
        """
        logging hyper parameters

        Args:
            dic (Dict):
            prefix (str):

        Returns:
            None

        """
        self.experiment.log_parameters(self.optimizer.__dict__["defaults"], prefix="optimizer")
        self.experiment.log_parameter("Optimizer", self.optimizer.__class__.__name__)

        if dic is not None:
            self.experiment.log_parameters(dic, prefix)

    def initialize(self) -> None:
        """
        The method that prepares the Runner.
        If the variables required for execution, such as self.model, self.optimizer, self.experiment, etc.,
        are not defined, the program will exit with an error message.
        If there are no problems, pass the model to the CPU or GPU.

        Returns:
            None

        """

        if not isinstance(self.model, Module):
            raise NotImplementedError("`self.model` is not defined.")

        if not isinstance(self.optimizer, Optimizer):
            raise NotImplementedError("`self.optimizer` is not defined.")

        if self.experiment is NotImplemented:
            raise NotImplementedError("`self.experiment` is not defined.")

        if self.scheduler and not isinstance(self.scheduler, list):
            raise ValueError("`scheduler` must be a list object.")

        if isinstance(self.experiment, Experiment):
            self.api_experiment = APIExperiment(previous_experiment=self.experiment.id, cache=False)

        if self.global_step < 0:
            self.global_step = 0

        self.manager.callbacks = self.callbacks

        self.manager.set_experiment(self.experiment)
        self.manager.set_device(self.device)
        self.manager.set_optimizer(self.optimizer)
        self.manager.set_model(self.model)

        self.model = self.model.to(self.device)
        self.save_dir = self.configures["checkpoint_path"]

    def run(self, phase: str = "all", verbose: bool = True, sleep_time: int = 1):
        """
        Runners are executed.
        To run it, you must register a data loader using self.add_loader() before.

        Args:
            phase (str):
                - `train`
                - `val`
                - `test`
                - `all`
                - `debug`

                by specifying one of the above, you can determine the execution phase. Default: all

            verbose (bool): If true, progress is displayed.
            sleep_time (int): The time to wait for data transfer to the comet.ml server (in seconds). Default: 1 (sec).

        Notes:
            If "TypeError" occurs even though there is no mistake in the formula of the monitor specified by
            `.train_config()`, the reason may be that it takes a long time to transfer the data to the comet.ml server.
            Try to set the `sleep_time` to about 5 seconds.

        Returns:
            None

        """
        phases = {"train", "train/val", "test", "all", "debug"}
        if phase not in phases:
            raise KeyError("The argument 'phase' must be one of the following. {}".format(phases))

        if phase == "debug":
            if hasattr(self.experiment, "add_tag"):
                self.experiment.add_tag("debug")  # type: ignore

        self.initialize()
        self.log_hyperparams()

        if not self.loaders:
            raise ValueError("At least one DataLoader must be provided.")

        if phase in {"all", "train", "train/val", "debug"}:
            if "train" in self.loaders:
                self.pbar = (
                    tqdm(range(self.configures["epochs"]), desc="Epochs")
                    if verbose
                    else range(self.configures["epochs"])
                )

                for epoch in self.pbar:
                    self.manager.on_epoch_start(epoch, self.metrics)

                    self.manager.on_train_start(self.metrics)
                    self.train_cycle(epoch, self.loaders["train"])
                    self.manager.on_train_end(self.metrics)

                    if phase in {"all", "train/val", "debug"}:
                        if "val" in self.loaders:
                            self.manager.on_validation_start(self.metrics)
                            self.val_cycle(epoch, self.loaders["val"])
                            self.manager.on_validation_end(self.metrics)

                    if self.scheduler:
                        self.update_scheduler(epoch)

                    self.manager.on_epoch_end(epoch, self.metrics)

                    if self.manager.stop_runner:
                        if self.manager.params["best_weight"]:
                            self.model = restore_state_dict(self.model, self.manager.params["best_weight"])
                        if hasattr(self.pbar, "close"):
                            self.pbar.close()  # type: ignore
                        break

                    if self.configures["checkpoint_path"]:
                        ops = {
                            "==": operator.eq,
                            "!=": operator.ne,
                            "<": operator.lt,
                            "<=": operator.le,
                            ">": operator.gt,
                            ">=": operator.ge,
                        }

                        if "monitor" in self.configures.keys():
                            key, op, value = self.configures["monitor"].split(" ")
                            value = float(value)
                            sleep(sleep_time)

                            try:
                                current_value = float(
                                    self.api_experiment.get_metrics_summary(key)["valueCurrent"]  # type: ignore
                                )
                            except TypeError:
                                raise KeyError(
                                    "The specified key was not found. Check the settings of `.train_config()`."
                                )

                            if current_value:
                                if ops[op](current_value, value):
                                    self.save(self.configures["checkpoint_path"], epoch=epoch)

                        else:
                            self.save(self.configures["checkpoint_path"], epoch=epoch)

        if phase in {"all", "test", "debug"}:
            if "test" in self.loaders:
                self.manager.on_test_start(self.metrics)
                self.pbar = tqdm(total=len(self.loaders["test"]), desc="Evaluating") if verbose else None
                self.test_cycle(self.loaders["test"])
                self.manager.on_test_end(self.metrics)

        return self

    def predict(self, x: Union[Tensor, ndarray]) -> ndarray:
        """
        A method that makes predictions based on the given input.

        Args:
            x (Union[torch.Tensor, np.ndarray]):

        Returns:
            predict
        """
        raise NotImplementedError

    def add_loader(self, mode: str, loader: Union[DataLoader, Any]):
        """
        A method to register a DataLoader to be used for training etc. in a runner.

        Args:
            mode (str): Specify one of ['train', 'val', 'test'].
            loader (torch.utils.data.DataLoader):

        Examples:
            >>> train_loader = DataLoader(...)
            >>> runner: BaseRunner = ...
            >>> runner.add_loader("train", train_loader)

        """
        if mode not in ["train", "val", "test"]:
            raise KeyError("argument `mode` must be one of 'train', 'val', or 'test'.")

        if is_tfds(loader) or isinstance(loader, DataLoader):
            pass

        else:
            raise TypeError("The argument `loader` must be an instance of `torch.utils.data.DataLoader`.")

        self._loaders[mode] = loader
        return self

    @property
    def loaders(self) -> Dict[str, DataLoader]:
        if len(self._loaders) != 0:
            return self._loaders
        else:
            raise ValueError

    def fit(self, x: ndarray, y: ndarray, **kwargs):
        """
        Scikit-Learn style training method.

        Args:
            x: Training data
            y: Label
            **kwargs:

        """
        val_size: float = kwargs.get("val_size", 0.1)
        num_workers: int = kwargs.get("num_workers", os.cpu_count())
        batch_size: int = kwargs.get("batch_size", 1)
        pin_memory: bool = kwargs.get("pin_memory", False)
        verbose: bool = kwargs.get("verbose", True)
        checkpoint_path = kwargs.get("checkpoint_path", None)
        monitor = kwargs.get("monitor", None)

        if self.configures["epochs"] == 0:
            epochs = kwargs.get("epochs", 1)
        else:
            epochs = self.configures["epochs"]

        train_ds = get_dataset(x, y)
        val_ds = get_dataset(x, y)
        n_train = len(train_ds)
        indices = list(range(n_train))
        split = int(floor(val_size * n_train))

        train_idx, val_idx = indices[split:], indices[:split]
        train_sampler = SubsetRandomSampler(train_idx)
        val_sampler = SubsetRandomSampler(val_idx)

        train_loader = DataLoader(
            train_ds,
            batch_size=batch_size,
            sampler=train_sampler,
            num_workers=num_workers,
            pin_memory=pin_memory,
        )
        val_loader = DataLoader(
            val_ds,
            batch_size=batch_size,
            sampler=val_sampler,
            num_workers=num_workers,
            pin_memory=pin_memory,
        )

        self.add_loader("train", train_loader)
        self.add_loader("val", val_loader)
        self.train_config(epochs, checkpoint_path=checkpoint_path, monitor=monitor)
        self.run(verbose=verbose)

        return self

    def freeze(self) -> None:
        """
        A method to freeze the model's parameters so that they do not calculate the slope.

        """
        for param in self.model.parameters():
            param.requires_grad = False

        self.model.eval()

    def unfreeze(self) -> None:
        """
        The method to make it possible to re-learn the parameters fixed by `.freeze()`.

        """
        for param in self.model.parameters():
            param.requires_grad = True

        self.model.train()

    def quite(self) -> None:
        """
        Quit Runner.

        When this method is executed, it sends an exit command to `comet.ml`.

        """

        self.experiment.end()

    def __enter__(self):
        """
        Context API

        Examples:
            >>> runner: BaseRunner = ...
            >>> with runner:
            >>>     runner.add_loader(...)
            >>>     runner.train_config(...)
            >>>     runner.run()

        """
        self.initialize()
        self.log_hyperparams()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        At the end of the ``with`` block, the Experiment will end automatically.

        """
        buffer = io.BytesIO()
        torch.save(self.save_checkpoint(), buffer)
        self.experiment.log_asset_data(
            buffer.getvalue(),
            step=self.global_step,
            name="context_api/enchanter_checkpoints_latest.pth",
        )
        self.quite()



from .loggers import BaseLogger, TensorBoardLogger
from .early_stopping import EarlyStopping, EarlyStoppingForTSUS
from .manager import CallbackManager
from .base import Callback

__all__ = ["TensorBoardLogger", "BaseLogger", "EarlyStopping", "CallbackManager", "Callback", "EarlyStoppingForTSUS"]



from typing import Callable, Optional

import torch
from torch.utils.data import Dataset


class TimeSeriesUnlabeledDataset(Dataset):
    def __init__(self, data: torch.Tensor, transform: Optional[Callable[[torch.Tensor], torch.Tensor]] = None):
        self.data: torch.Tensor = data
        self.transform = transform

    def __len__(self):
        return self.data.shape[0]

    def __getitem__(self, item):
        data = self.data[item]
        if self.transform:
            data = self.transform(data)

        return data


class TimeSeriesLabeledDataset(TimeSeriesUnlabeledDataset):
    def __init__(
        self,
        data: torch.Tensor,
        target: torch.Tensor,
        transform: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
    ):
        super(TimeSeriesLabeledDataset, self).__init__(data, transform)
        self.target = target

    def __getitem__(self, item):
        data = super(TimeSeriesLabeledDataset, self).__getitem__(item)
        target = self.target[item]
        return data, target



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from typing import Tuple, List, Union, Optional, Dict

import numpy as np
import torch
from torch.nn.modules import Module
from torch.nn.modules.loss import _Loss
from torch.optim.optimizer import Optimizer
from torch.cuda.amp import GradScaler, autocast
from comet_ml.experiment import BaseExperiment

from enchanter.engine import BaseRunner
from enchanter.callbacks import BaseLogger
from enchanter.callbacks import Callback
from enchanter.metrics import calculate_accuracy


__all__ = ["ClassificationRunner"]


class ClassificationRunner(BaseRunner):
    """
    Runner for classification tasks.

    Examples:
        >>> from comet_ml import Experiment
        >>> import torch
        >>> model = torch.nn.Sequential(...)
        >>> optimizer = torch.optim.Adam(model.parameters())
        >>> criterion = torch.nn.CrossEntropyLoss()
        >>> runner = ClassificationRunner(
        >>>     model,
        >>>     optimizer,
        >>>     criterion,
        >>>     Experiment()
        >>> )
        >>> runner.fit(...)
        >>> # or
        >>> runner.add_loader(...)
        >>> runner.train_config(epochs=10)
        >>> runner.run()

    """

    def __init__(
        self,
        model: Module,
        optimizer: Optimizer,
        criterion: _Loss,
        experiment: Union[BaseExperiment, BaseLogger],
        scheduler: Optional[List] = None,
        callbacks: Optional[List[Callback]] = None,
    ) -> None:
        super(ClassificationRunner, self).__init__()
        self.model: Module = model
        self.optimizer: Optimizer = optimizer
        self.experiment: Union[BaseExperiment, BaseLogger] = experiment
        self.criterion: _Loss = criterion
        if scheduler is None:
            self.scheduler: List = list()
        else:
            self.scheduler = scheduler

        self.callbacks = callbacks

    def general_step(self, batch: Tuple) -> Dict:
        """
        This method is executed by train_step, val_step, test_step.

        Args:
            batch:

        Returns:

        """
        x, y = batch

        with autocast(enabled=isinstance(self.scaler, GradScaler)):
            out = self.model(x)
            loss = self.criterion(out, y)

        accuracy = calculate_accuracy(out, y)
        return {"loss": loss, "accuracy": accuracy}

    @staticmethod
    def general_end(outputs: List) -> Dict:
        """
        This method is executed by train_end, val_end, test_end.

        Args:
            outputs:

        Returns:

        """
        avg_loss = torch.stack([x["loss"] for x in outputs]).mean()
        avg_acc = torch.stack([torch.tensor(x["accuracy"]) for x in outputs]).mean()
        return {"avg_loss": avg_loss, "avg_acc": avg_acc}

    def train_step(self, batch: Tuple) -> Dict:
        return self.general_step(batch)

    def train_end(self, outputs: List) -> Dict:
        return self.general_end(outputs)

    def val_step(self, batch: Tuple) -> Dict:
        return self.general_step(batch)

    def val_end(self, outputs: List) -> Dict:
        return self.general_end(outputs)

    def test_step(self, batch: Tuple) -> Dict:
        return self.general_step(batch)

    def test_end(self, outputs: List) -> Dict:
        return self.general_end(outputs)

    def predict(self, x: Union[torch.Tensor, np.ndarray]) -> np.ndarray:
        self.model.eval()
        with torch.no_grad():
            x = torch.as_tensor(x, device=self.device)
            out = self.model(x)
            _, predicted = torch.max(out, 1)

        return predicted.cpu().numpy()



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from collections import Counter
from typing import Union, List, Type, Optional, Callable

from numpy import array, stack, ndarray, max as np_max, zeros, nan, float32, dstack
from pandas import DataFrame
from tqdm.auto import tqdm


__all__ = ["FixedSlidingWindow", "adjust_sequences"]


_Numerical = Union[int, float]


class FixedSlidingWindow:
    """Fixed sliding window.

    Examples::
        >>> import numpy as np
        >>> from enchanter.preprocessing import signal
        >>> x = np.random.randn(1024, 23)
        >>> y = np.random.randint(0, 9, 1024)
        >>> sw = signal.FixedSlidingWindow(256, overlap_rate=0.5)
        >>> x, y = sw(x, y)
        >>> x.shape     # [6, 256, 23]
        >>> y.shape     # [6, ]

    Args:
        window_size (int): Window size
        overlap_rate (float): overrap rate
        step_size (Optional[int]): step size

    Raises:
        AssertionError: an error occur when
            argument overlap_rate under 0.0 or over 1.0.n error occurred.

    """

    def __init__(self, window_size: int, overlap_rate: Union[float, None], step_size: Optional[int] = None) -> None:
        self.window_size: int = window_size

        if overlap_rate is None and step_size is not None:
            if step_size > 0:
                self.overlap = int(step_size)
        elif isinstance(overlap_rate, float):
            if not 0.0 < overlap_rate <= 1.0:
                raise AssertionError("overlap_rate ranges from 0.0 to 1.0")

            self.overlap = int(window_size * overlap_rate)
        else:
            raise ValueError

    def transform(self, inputs: ndarray, verbose: bool = False) -> ndarray:
        """
        Apply Fixed Sliding Window

        Args:
            inputs: 2 or 3 dim of np.ndarray
            verbose: if True, show progress bar

        Returns:
            np.ndarray

        """
        seq_len = inputs.shape[0]
        if not seq_len > self.window_size:
            raise IndexError(
                "window size ({}) must be smaller then input sequence length ({}).".format(self.window_size, seq_len)
            )

        if verbose:
            data = []
            for i in tqdm(range(0, seq_len - self.window_size, self.overlap)):
                data.append(inputs[i : i + self.window_size])
        else:
            data = [inputs[i : i + self.window_size] for i in range(0, seq_len - self.window_size, self.overlap)]

        data = stack(data, 0)
        return data

    @staticmethod
    def clean(labels: ndarray) -> ndarray:
        """
        Clean up

        Args:
            labels:

        Returns:

        """
        tmp = []
        for lbl in labels:
            window_size = len(lbl)
            counter: Counter = Counter(lbl)
            common = counter.most_common()
            values = list(counter.values())
            if common[0][0] == 0 and values[0] == window_size // 2:
                label = common[1][0]
            else:
                label = common[0][0]

            tmp.append(label)

        return array(tmp)

    def __call__(self, data, target):
        data = self.transform(data)
        label = self.transform(target)
        label = self.clean(label)
        return data, label


def adjust_sequences(
    sequences: List[ndarray],
    max_len: Optional[Union[int, Callable[[List[int]], int]]] = None,
    fill: Union[str, _Numerical] = "ffill",
    dtype: Type = float32,
) -> ndarray:
    """
    The function to adjust the length of the series data to a certain value.
    For each sample, if the series is longer than ``max_len``, the length is up to ``max_len``, and the rest is ignored.
    If it is shorter than ``max_len``, the missing part is filled in with the last value.

    Args:
        sequences: A Python list whose elements have ``np.ndarray`` objects that are not constant in length.
                   Each element is a 2D array, the 0th dimension is the length of the series, the 1st dimension is
                   the number of features in the time series, and all samples must have the same number of features.

        max_len: Processes all input elements to the specified length.
                    If not specified, the largest sequence length in a given sample will be max_len.
                    Also, given functions such as np.max, np.min, and np.mean,
                    you can use them to generate new length series.

        fill: If it is shorter than ``max_len``, specify how to fill in the missing parts.
                If ``fill ='ffill'``, it will be filled with the last value.
                If a number (Python int or Python float) is given, use that value to fill the value.
                ``fill=["ffill" or int or float]``

        dtype:  Specify the data type of NumPy. The data type of the output series is determined based on this value.

    Examples:
        >>> import numpy as np
        >>> x = [
        >>>         np.array([[i] for i in [1, 2, 3, 4, 5]]),
        >>>         np.array([[i] for i in [1, 2, 3, 4, 5, 6, 7, 8]]),
        >>>         np.array([[i] for i in [1, 2, 3]]),
        >>> ]
        >>> out = adjust_sequences(x)
        >>> out[-1]
        >>> # array([[1.],
        >>> #        [2.],
        >>> #        [3.],
        >>> #        [3.],
        >>> #        [3.],
        >>> #        [3.],
        >>> #        [3.],
        >>> #        [3.]])
        >>> out = adjust_sequences(x, np.min)
        >>> out
        >>> # array([[[1],
        >>> #         [2],
        >>> #         [3]],
        >>> #
        >>> #        [[1],
        >>> #         [2],
        >>> #         [3]],
        >>> #
        >>> #        [[1],
        >>> #         [2],
        >>> #         [3]]])

    Returns:
        Returns a 3D array of ``[Sample, Seq_len, Features]``.

    """
    features = sequences[0].shape[1]

    lengths = []
    for item in sequences:
        if isinstance(item, ndarray):
            lengths.append(item.shape[0])

    if max_len is None:
        maximum_len = np_max(lengths)
    elif callable(max_len):
        maximum_len = int(max_len(lengths))
    else:
        maximum_len = max_len

    new_seqs = []
    for seq in sequences:
        new_seq = zeros((maximum_len, features), dtype=dtype)
        new_seq[:, :] = nan
        new_seq = DataFrame(new_seq)

        if seq.dtype != dtype:
            seq = seq.astype(dtype)

        if maximum_len > seq.shape[0]:
            new_seq[: seq.shape[0]] = seq
            if fill == "ffill":
                new_seq = new_seq.ffill()
            elif isinstance(fill, int) or isinstance(fill, float):
                new_seq = new_seq.fillna(fill)
            else:
                raise TypeError
        else:
            new_seq[:maximum_len] = seq[:maximum_len]

        new_seqs.append(new_seq.values)

    return dstack(new_seqs).transpose((2, 0, 1))



from .ts_triplet_loss import (
    generate_sample_indices,
    generate_anchor_positive_input,
    generate_negative_input,
    positive_criterion_for_triplet_loss,
    negative_criterion_for_triplet_loss,
    calculate_triplet_loss,
)


__all__ = [
    "generate_sample_indices",
    "generate_anchor_positive_input",
    "generate_negative_input",
    "positive_criterion_for_triplet_loss",
    "negative_criterion_for_triplet_loss",
    "calculate_triplet_loss",
]



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from typing import Any, Optional, Dict, Union, Iterator, List
from abc import ABC, abstractmethod
from contextlib import contextmanager

import numpy as np
from torch import Tensor
from torch.utils.tensorboard import SummaryWriter


__all__ = ["BaseLogger", "TensorBoardLogger"]


class BaseLogger(ABC):
    """
    Provides minimal compatibility with the `comet_ml.Experiment`, which is required to run Runner.

    """

    def __init__(self) -> None:
        self.context: Union[str, None] = None

    def add_tag(self, tag: str):
        pass

    def add_tags(self, tags: List[str]):
        pass

    @contextmanager
    def train(self) -> Iterator:
        old_state = self.context
        self.context = "train"

        yield self

        self.context = old_state

    @contextmanager
    def validate(self) -> Iterator:
        old_state = self.context
        self.context = "validate"

        yield self

        self.context = old_state

    @contextmanager
    def test(self) -> Iterator:
        old_state = self.context
        self.context = "test"

        yield self

        self.context = old_state

    @abstractmethod
    def log_metric(
        self,
        name: str,
        value: Any,
        step: Optional[int] = None,
        epoch: Optional[int] = None,
        include_context: bool = True,
    ) -> None:
        """
        Logs a metric.

        Args:
            name: name of metric
            value:
            step:
            epoch:
            include_context:

        Returns:
            None

        Warnings:
            If you create your own Logger, you will need to implement this method.

        """
        raise NotImplementedError

    @abstractmethod
    def log_metrics(
        self,
        dic: Dict,
        prefix: Optional[str] = None,
        step: Optional[int] = None,
        epoch: Optional[int] = None,
    ) -> None:
        """

        Logs a key, value dictionary of metrics.

        See Also:
            log_metric


        Warnings:
            If you create your own Logger, you will need to implement this method.

        """
        raise NotImplementedError

    @abstractmethod
    def log_parameter(self, name: str, value: Any, step: Optional[int] = None) -> None:
        """
        Logs a single hyper-parameter.

        Args:
            name: name of hyper-parameter
            value: value
            step:

        Warnings:
            If you create your own Logger, you will need to implement this method.

        """
        raise NotImplementedError

    @abstractmethod
    def log_parameters(self, dic: Dict, prefix: Optional[str] = None, step: Optional[int] = None) -> None:
        """
        Logs a key, value dictionary of hyper-parameters.

        See Also:
            log_peramter


        Warnings:
            If you create your own Logger, you will need to implement this method.

        """
        raise NotImplementedError

    def set_model_graph(self, *args: Any, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def end(self):
        """

        Use to indicate that the experiment is complete.

        Warnings:
            If you create your own Logger, you will need to implement this method.

        """
        raise NotImplementedError

    def log_model(self, name, file_or_folder, file_name=None, overwrite=False, metadata=None, copy_to_tmp=True):
        pass

    def log_asset(self, file_data, file_name=None, overwrite=False, copy_to_tmp=True, step=None, metadata=None):
        pass

    def log_asset_data(self, data, name=None, overwrite=False, step=None, metadata=None, file_name=None):
        pass

    def log_asset_folder(self, folder, step=None, log_file_name=False, recursive=False):
        pass


def _value2str(value) -> str:
    if isinstance(value, Tensor):
        value = value.cpu().item()
    elif isinstance(value, np.ndarray):
        value = value.item()

    return str(value)


class TensorBoardLogger(BaseLogger):
    """
    TenorBoardLogger is a module that supports the minimum logging in the environment where comet.ml cannot be used.

    Examples:
        >>> from enchanter.tasks import ClassificationRunner
        >>> model, optimizer, criterion = ...
        >>> runner = ClassificationRunner(
        >>>     model, optimizer, criterion, experiment=TensorBoardLogger()
        >>> )

    """

    def __init__(self, *args, **kwargs):
        """
        See the initializer argument for `torch.utils.tensorboard.writer.SummaryWriter
        <https://pytorch.org/docs/master/tensorboard.html>`_ .
        """
        super(TensorBoardLogger, self).__init__()
        self.writer: SummaryWriter = SummaryWriter(*args, **kwargs)

    def log_metric(
        self,
        name: str,
        value: Any,
        step: Optional[int] = None,
        epoch: Optional[int] = None,
        include_context: bool = True,
    ) -> None:
        """
        Logs a metric.

        Args:
            name: name of metric
            value: value
            step: Used as the X axis when plotting on TensorBoard.
            epoch: Used as the X axis when plotting on TensorBoard.
            include_context:

        Returns:
            None

        """
        if step is not None:
            tmp = step
        else:
            if epoch is not None:
                tmp = epoch
            else:
                tmp = 0

        if step is not None and epoch is None:
            tmp = step
        elif step is None and epoch is not None:
            tmp = epoch

        if isinstance(value, list):
            value = np.array(value)

        self.writer.add_scalar("{}/{}".format(name, self.context), value, global_step=tmp)

    def log_metrics(
        self,
        dic: Dict,
        prefix: Optional[str] = None,
        step: Optional[int] = None,
        epoch: Optional[int] = None,
    ) -> None:
        """
        Logs a key, value dictionary of metrics.

        See Also:
            log_metric

        """
        for k, v in dic.items():
            if prefix:
                k = "{}_{}".format(prefix, k)
            self.log_metric(name=k, value=v, step=step, epoch=epoch)

    def log_parameter(self, name: Optional[str], value: Any, step: Optional[int] = None) -> None:
        """
        Logs a single hyper-parameter.

        Args:
            name: name of hyper-parameter
            value: value
            step: used as the X-axis when plotting on TensorBoard.

        Returns:

        """
        value = _value2str(value)
        if self.context:
            prefix = self.context
        else:
            prefix = "Hyperparameters"

        if name:
            table = "|key|value|\n|-|-|\n|{}|{}|  \n".format(name, value)
        else:
            table = value

        self.writer.add_text("{}/{}".format(prefix, name), table, global_step=step)

    def log_parameters(self, dic: Dict, prefix: Optional[str] = None, step: Optional[int] = None) -> Any:
        """
        Logs a key, value dictionary of hyper-parameters.

        See Also:
            log_peramter

        """
        table = "|key|value|\n|-|-|\n"
        for k, v in dic.items():
            table += "|{}|{}|  \n".format(k, v)

        self.log_parameter(None, table, step)

    def end(self):
        """
        Use to indicate that the experiment is complete.

        """
        self.writer.close()



from typing_extensions import Protocol


__all__ = ["ScikitModel"]


class ScikitModel(Protocol):
    # pylint: disable=R0201
    def fit(self, X, y, sample_weight=None):
        # pylint: disable=W0104
        ...

    # pylint: disable=R0201
    def predict(self, X):
        # pylint: disable=W0104
        ...

    # pylint: disable=R0201
    def score(self, X, y, sample_weight=None):
        # pylint: disable=W0104
        ...

    # pylint: disable=R0201
    def set_params(self, **params):
        # pylint: disable=W0104
        ...

    # pylint: disable=R0201
    def decision_function(self, X):
        # pylint: disable=W0104
        ...

    # pylint: disable=R0201
    def get_params(self, deep=True):
        # pylint: disable=W0104
        ...



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from abc import ABC
from typing import Dict, Optional


__all__ = ["Callback"]


class Callback(ABC):
    def __init__(self):
        self.stop_runner: bool = False
        self.model = None
        self.optimizer = None
        self.device = None
        self.experiment = None
        self.params = {}

    def set_device(self, device):
        self.device = device

    def set_model(self, model):
        self.model = model

    def set_optimizer(self, optimizer):
        self.optimizer = optimizer

    def on_epoch_start(self, epoch, logs: Optional[Dict] = None):
        """Called when the epoch begins."""
        pass

    def on_epoch_end(self, epoch, logs: Optional[Dict] = None):
        """Called when the epoch ends."""
        pass

    def on_train_step_start(self, logs: Optional[Dict] = None):
        """Called when the training batch begins."""
        pass

    def on_train_step_end(self, logs: Optional[Dict] = None):
        """Called when the training batch ends. You can access the output of train_step."""
        pass

    def on_validation_step_start(self, logs: Optional[Dict] = None):
        pass

    def on_validation_step_end(self, logs: Optional[Dict] = None):
        """Called when the validation batch ends. You can access the output of val_step."""
        pass

    def on_test_step_start(self, logs: Optional[Dict] = None):
        pass

    def on_test_step_end(self, logs: Optional[Dict] = None):
        """Called when the test batch ends. You can access the output of test_step."""
        pass

    def on_train_start(self, logs: Optional[Dict] = None):
        """Called when the train begins."""
        pass

    def on_train_end(self, logs: Optional[Dict] = None):
        """Called when the train ends."""
        pass

    def on_validation_start(self, logs: Optional[Dict] = None):
        """Called when the validation loop begins."""
        pass

    def on_validation_end(self, logs: Optional[Dict] = None):
        """Called when the validation loop ends."""
        pass

    def on_test_start(self, logs: Optional[Dict] = None):
        """Called when the test begins."""
        pass

    def on_test_end(self, logs: Optional[Dict] = None):
        """Called when the test ends."""
        pass



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from typing import Tuple, List, Union, Optional, Dict

import numpy as np
from sklearn.metrics import r2_score
import torch
from torch.nn.modules import Module
from torch.nn.modules.loss import _Loss
from torch.optim.optimizer import Optimizer
from torch.cuda.amp import GradScaler, autocast
from comet_ml.experiment import BaseExperiment

from enchanter.engine import BaseRunner
from enchanter.callbacks import BaseLogger
from enchanter.callbacks import Callback


__all__ = ["RegressionRunner", "AutoEncoderRunner"]


class RegressionRunner(BaseRunner):
    """
    Runner for regression problems.

    Examples:
        >>> runner = RegressionRunner(...)
        >>> runner.add_loader("train", ...)
        >>> runner.train_config(epochs=1)
        >>> runner.run()
        >>> # OR
        >>> runner = RegressionRunner(...)
        >>> runner.fit(x, y, epochs=1, batch_size=32)
    """

    def __init__(
        self,
        model: Module,
        optimizer: Optimizer,
        criterion: _Loss,
        experiment: Union[BaseExperiment, BaseLogger],
        scheduler: Optional[List] = None,
        callbacks: Optional[List[Callback]] = None,
    ) -> None:
        super(RegressionRunner, self).__init__()
        self.model: Module = model
        self.optimizer: Optimizer = optimizer
        self.criterion: _Loss = criterion
        self.experiment: Union[BaseExperiment, BaseLogger] = experiment
        if scheduler is None:
            self.scheduler: List = list()
        else:
            self.scheduler = scheduler
        self.callbacks = callbacks

    def general_step(self, batch: Tuple) -> Dict:
        """
        This method is executed by train_step, val_step, test_step.

        Args:
            batch:

        Returns:

        """
        x, y = batch

        with autocast(enabled=isinstance(self.scaler, GradScaler)):
            out = self.model(x)
            loss = self.criterion(out, y)

        r2 = r2_score(y.cpu().numpy(), out.cpu().detach().numpy())
        return {"loss": loss, "r2": r2}

    @staticmethod
    def general_end(outputs: List) -> Dict:
        """
        This method is executed by train_end, val_end, test_end.

        Args:
            outputs:

        Returns:

        """
        avg_loss = torch.stack([x["loss"] for x in outputs]).mean()
        avg_r2 = torch.stack([torch.tensor(x["r2"]) for x in outputs]).mean()
        return {"avg_loss": avg_loss, "avg_r2": avg_r2}

    def train_step(self, batch: Tuple) -> Dict:
        return self.general_step(batch)

    def train_end(self, outputs: List) -> Dict:
        return self.general_end(outputs)

    def val_step(self, batch: Tuple) -> Dict:
        return self.general_step(batch)

    def val_end(self, outputs: List) -> Dict:
        return self.general_end(outputs)

    def test_step(self, batch: Tuple) -> Dict:
        return self.general_step(batch)

    def test_end(self, outputs: List) -> Dict:
        return self.general_end(outputs)

    def predict(self, x: Union[torch.Tensor, np.ndarray]) -> np.ndarray:
        self.model.eval()
        with torch.no_grad():
            x = torch.as_tensor(x, device=self.device)
            out = self.model(x)

        return out.cpu().numpy()


class AutoEncoderRunner(RegressionRunner):
    """
    Runner for training AutoEncoder.

    """

    def general_step(self, batch: Tuple) -> Dict:
        """
        This method is executed by train_step, val_step, test_step.

        Args:
            batch:

        Returns:

        """
        x, _ = batch
        with autocast(enabled=isinstance(self.scaler, GradScaler)):
            out = self.model(x)
            loss = self.criterion(out, x)

        return {"loss": loss}

    @staticmethod
    def general_end(outputs: List) -> Dict:
        """
        This method is executed by train_end, val_end, test_end.

        Args:
            outputs:

        Returns:

        """
        avg_loss = torch.stack([x["loss"] for x in outputs]).mean()
        return {"avg_loss": avg_loss}



from typing import List, Optional, Dict
from .base import Callback


__all__ = ["CallbackManager"]


class CallbackManager(Callback):
    """
    A class for managing Callback.

    Examples:
        >>> from enchanter.callbacks import EarlyStopping
        >>> manager = CallbackManager([
        >>>     EarlyStopping()
        >>> ])
        >>> for epoch in range(5):
        >>>     logs = ...
        >>>     manager.on_epoch_end(epoch, logs)
        >>>     if manager.stop_runner:
        >>>         break

    """

    def __init__(self, callbacks: Optional[List[Callback]] = None):
        super(CallbackManager, self).__init__()
        self.callbacks = callbacks

    def set_experiment(self, experiment):
        self.experiment = experiment
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.experiment = self.experiment

    def set_device(self, device):
        self.device = device
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.device = self.device

    def set_model(self, model):
        self.model = model
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.model = self.model

    def set_optimizer(self, optimizer):
        self.optimizer = optimizer
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.optimizer = self.optimizer

    def flag_check(self, stop_runner: bool):
        if stop_runner and self.callbacks is not None:
            self.stop_runner = stop_runner

    def on_epoch_start(self, epoch, logs=None):
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.on_epoch_start(epoch, logs)

                self.flag_check(callback.stop_runner)

    def on_epoch_end(self, epoch, logs=None):
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.on_epoch_end(epoch, logs)

                self.flag_check(callback.stop_runner)
                if "best_weight" in callback.params.keys():
                    self.params["best_weight"] = callback.params["best_weight"]

    def on_train_step_start(self, logs: Optional[Dict] = None):
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.on_train_step_start(logs)

                self.flag_check(callback.stop_runner)

    def on_train_step_end(self, logs: Optional[Dict] = None):
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.on_train_step_end(logs)

                self.flag_check(callback.stop_runner)

    def on_validation_step_start(self, logs: Optional[Dict] = None):
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.on_validation_step_start(logs)

                self.flag_check(callback.stop_runner)

    def on_validation_step_end(self, logs: Optional[Dict] = None):
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.on_validation_step_end(logs)

                self.flag_check(callback.stop_runner)

    def on_test_step_start(self, logs: Optional[Dict] = None):
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.on_test_step_start(logs)

                self.flag_check(callback.stop_runner)

    def on_test_step_end(self, logs: Optional[Dict] = None):
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.on_test_step_end(logs)

                self.flag_check(callback.stop_runner)

    def on_train_start(self, logs=None):
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.on_train_start(logs)

                self.flag_check(callback.stop_runner)

    def on_train_end(self, logs=None):
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.on_train_end(logs)

                self.flag_check(callback.stop_runner)

    def on_validation_start(self, logs=None):
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.on_validation_start(logs)

                self.flag_check(callback.stop_runner)

    def on_validation_end(self, logs=None):
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.on_validation_end(logs)

                self.flag_check(callback.stop_runner)

    def on_test_start(self, logs=None):
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.on_test_start(logs)

                self.flag_check(callback.stop_runner)
                if "grid_search" in callback.params.keys():
                    self.params["grid_search"] = callback.params["grid_search"]

    def on_test_end(self, logs=None):
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.on_test_end(logs)

                self.flag_check(callback.stop_runner)



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

import warnings
from copy import deepcopy
from typing import Union, Tuple, Any, Dict
from os import environ as os_environ
from random import seed as std_seed

from numpy import ndarray
from numpy.random import seed as np_seed
import torch
import torch.nn as nn
from torch.backends import cudnn
from torch.utils.data import Dataset, TensorDataset
from torch.cuda import is_available as cuda_is_available

try:
    import tensorflow as tf

    IS_TF_DS_AVAILABLE = True

except ImportError:
    IS_TF_DS_AVAILABLE = False


__all__ = [
    "is_jupyter",
    "get_dataset",
    "fix_seed",
    "send",
    "is_tfds",
    "tfds_to_numpy",
    "fetch_state_dict",
    "restore_state_dict",
]


def is_jupyter() -> bool:
    """
    Determines if the running environment is a Jupyter Notebook.

    Returns:
        True if run on jupyrer notebook else False

    """
    try:
        from IPython import get_ipython
    except ImportError:
        return False

    env = get_ipython().__class__.__name__  # noqa
    if env == "TerminalInteractiveShell":
        return False
    return True


def get_dataset(x: Union[ndarray, torch.Tensor], y: Union[ndarray, torch.Tensor] = None) -> Dataset:
    """
    Generates ``torch.utils.data.TensorDataset`` based on the values entered.

    Examples:
        >>> import torch
        >>> x = torch.randn(512, 6)
        >>> y = torch.randint(0, 9, size=[512])
        >>> ds = get_dataset(x, y)

    Args:
        x (Union[np.ndarray, torch.Tensor]):
        y (Optional[Union[np.ndarray, torch.Tensor]]):

    Returns:
        `torch.utils.data.TensorDataset`
    """
    x = torch.as_tensor(x)

    if y is not None:
        y = torch.as_tensor(y)
        ds = TensorDataset(x, y)
    else:
        ds = TensorDataset(x)

    return ds


def send(batch: Tuple[Any, ...], device: torch.device, non_blocking: bool = True) -> Tuple[Any, ...]:
    """
    Send `variable` to `device`

    Args:
        batch: Tuple which contain variable
        device: torch.device
        non_blocking: bool

    Returns:
        new tuple

    """

    def transfer(x):
        if isinstance(x, torch.Tensor):
            return x.to(device, non_blocking=non_blocking)
        elif isinstance(x, ndarray):
            return torch.tensor(x, device=device)
        else:
            return x

    return tuple(map(transfer, batch))


def fix_seed(seed: int, deterministic: bool = False, benchmark: bool = False) -> None:
    """
    Fixed the ``Seed`` value of PyTorch, NumPy, Pure Python Random at once.

    Examples:
        >>> import torch
        >>> import numpy as np
        >>> fix_seed(0)
        >>> x = torch.randn(...)
        >>> y = np.random.randn(...)

    Args:
        seed (int): random state (sedd)
        deterministic (bool): Whether to ensure reproducibility as much as possible on CuDNN.
        benchmark (bool):

    Returns:
        None
    """
    std_seed(seed)
    os_environ["PYTHONHASHSEED"] = str(seed)
    np_seed(seed)
    torch.manual_seed(seed)

    if cuda_is_available():
        if deterministic:
            cudnn.deterministic = True
        if benchmark:
            cudnn.benchmark = False


def is_tfds(loader: Any) -> bool:
    """
    Determines if the input is a TensorFlow Dataset. Returns False if TensorFlow is not installed.

    Args:
        loader:

    Returns: True if input is TensorFlow Dataset

    """
    if IS_TF_DS_AVAILABLE:
        if isinstance(loader, tf.data.Dataset):
            warnings.warn(
                "TensorFlow Dataset detection. Experimental support at this stage.",
                UserWarning,
            )
            return True

        else:
            return False
    else:
        return False


def tfds_to_numpy(loader):
    """
    If TensorFlow Dataset is entered, it will be converted to numpy.

    Args:
        loader:

    Returns: TensorFlow Dataset

    Raises: If `tfds <https://www.tensorflow.org/datasets>`_ is not installed, this function will raise a ``RuntimeError``.

    """
    if IS_TF_DS_AVAILABLE:
        return loader.as_numpy_iterator()
    else:
        raise RuntimeError


def fetch_state_dict(model: nn.Module) -> Dict[str, torch.Tensor]:
    if isinstance(model, nn.parallel.DataParallel) or isinstance(model, nn.parallel.DistributedDataParallel):
        weights = model.module.state_dict()
    else:
        weights = model.state_dict()

    return deepcopy(weights)


def restore_state_dict(model: nn.Module, weights: Dict[str, torch.Tensor], strict: bool = True) -> nn.Module:
    if isinstance(model, nn.parallel.DataParallel) or isinstance(model, nn.parallel.DistributedDataParallel):
        model.module.load_state_dict(weights, strict=strict)
    else:
        model.load_state_dict(weights, strict=strict)

    return model



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['enchanter',
 'enchanter.addons',
 'enchanter.addons.criterions',
 'enchanter.addons.layers',
 'enchanter.callbacks',
 'enchanter.engine',
 'enchanter.metrics',
 'enchanter.preprocessing',
 'enchanter.preprocessing.signal',
 'enchanter.tasks',
 'enchanter.utils']

package_data = \
{'': ['*']}

install_requires = \
['comet_ml>=3.1.3,<4.0.0',
 'netron>=4.5,<5.0',
 'numpy>=1.17,<2.0',
 'pandas>=1.0,<2.0',
 'scikit-learn>=0.23,<0.24',
 'tensorboard>=2.1,<3.0',
 'torch>=1.6,<2.0',
 'tqdm>=4.41,<5.0',
 'typing-extensions>=3.7.4,<4.0.0']

setup_kwargs = {
    'name': 'enchanter',
    'version': '0.8.0a0',
    'description': 'Enchanter is a library for machine learning tasks for comet.ml users.',
    'long_description': '<div align="center">\n\n# Enchanter\n\n<p align="center">\n  <a href="#Getting-Started">Getting Started</a> •\n  <a href="https://enchanter.readthedocs.io/en/stable/">Docs</a> •\n  <a href="https://enchanter.readthedocs.io/en/stable/tutorial/modules.html">Tutorial</a> •\n  <a href="LICENSE">Licence</a>\n</p>\n\n[![Codacy Badge](https://api.codacy.com/project/badge/Grade/84197fb283924f02a1667cea49dd031a)](https://app.codacy.com/manual/KawashimaHirotaka/enchanter?utm_source=github.com&utm_medium=referral&utm_content=khirotaka/enchanter&utm_campaign=Badge_Grade_Dashboard)\n[![CI testing](https://github.com/khirotaka/enchanter/workflows/CI/badge.svg)](https://github.com/khirotaka/enchanter/actions)\n[![Build & Publish](https://github.com/khirotaka/enchanter/workflows/Build%20&%20Publish/badge.svg)](https://github.com/khirotaka/enchanter/actions)\n[![Documentation Status](https://readthedocs.org/projects/enchanter/badge/?version=latest)](https://enchanter.readthedocs.io/)\n\n[![PyPI](https://img.shields.io/pypi/v/enchanter?color=brightgreen)](https://pypi.org/project/enchanter/)\n[![Dependabot Status](https://api.dependabot.com/badges/status?host=github&repo=khirotaka/enchanter)](https://dependabot.com)\n[![license](https://img.shields.io/github/license/khirotaka/enchanter?color=light)](LICENSE)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Using PyTorch](https://img.shields.io/badge/PyTorch-red.svg?labelColor=f3f4f7&logo=data:image/svg+xml;base64,PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0iVVRGLTgiPz4KPHN2ZyB3aWR0aD0iMjU2cHgiIGhlaWdodD0iMzEwcHgiIHZpZXdCb3g9IjAgMCAyNTYgMzEwIiB2ZXJzaW9uPSIxLjEiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyIgeG1sbnM6eGxpbms9Imh0dHA6Ly93d3cudzMub3JnLzE5OTkveGxpbmsiIHByZXNlcnZlQXNwZWN0UmF0aW89InhNaWRZTWlkIj4KICAgIDxnPgogICAgICAgIDxwYXRoIGQ9Ik0yMTguMjgxMDM3LDkwLjEwNjQxMiBDMjY4LjU3Mjk4OCwxNDAuMzk4MzYzIDI2OC41NzI5ODgsMjIxLjA3NTAzNCAyMTguMjgxMDM3LDI3MS43MTYyMzUgQzE2OS4wMzY4MzUsMzIyLjAwODE4NiA4OC4wMTA5MTQxLDMyMi4wMDgxODYgMzcuNzE4OTYzMiwyNzEuNzE2MjM1IEMtMTIuNTcyOTg3NywyMjEuNDI0Mjg0IC0xMi41NzI5ODc3LDE0MC4zOTgzNjMgMzcuNzE4OTYzMiw5MC4xMDY0MTIgTDEyNy44MjUzNzUsMCBMMTI3LjgyNTM3NSw0NS4wNTMyMDYgTDExOS40NDMzODMsNTMuNDM1MTk3OCBMNTkuNzIxNjkxNywxMTMuMTU2ODg5IEMyMi4wMDI3Mjg1LDE1MC4xNzczNTMgMjIuMDAyNzI4NSwyMTAuOTQ2Nzk0IDU5LjcyMTY5MTcsMjQ4LjY2NTc1NyBDOTYuNzQyMTU1NSwyODYuMzg0NzIgMTU3LjUxMTU5NiwyODYuMzg0NzIgMTk1LjIzMDU1OSwyNDguNjY1NzU3IEMyMzIuOTQ5NTIzLDIxMS42NDUyOTMgMjMyLjk0OTUyMywxNTAuODc1ODUzIDE5NS4yMzA1NTksMTEzLjE1Njg4OSBMMjE4LjI4MTAzNyw5MC4xMDY0MTIgWiBNMTczLjIyNzgzMSw4NC41MTg0MTc1IEMxNjMuOTY5MzM4LDg0LjUxODQxNzUgMTU2LjQ2Mzg0Nyw3Ny4wMTI5MjYzIDE1Ni40NjM4NDcsNjcuNzU0NDMzOCBDMTU2LjQ2Mzg0Nyw1OC40OTU5NDEzIDE2My45NjkzMzgsNTAuOTkwNDUwMiAxNzMuMjI3ODMxLDUwLjk5MDQ1MDIgQzE4Mi40ODYzMjMsNTAuOTkwNDUwMiAxODkuOTkxODE0LDU4LjQ5NTk0MTMgMTg5Ljk5MTgxNCw2Ny43NTQ0MzM4IEMxODkuOTkxODE0LDc3LjAxMjkyNjMgMTgyLjQ4NjMyMyw4NC41MTg0MTc1IDE3My4yMjc4MzEsODQuNTE4NDE3NSBaIiBmaWxsPSIjRUU0QzJDIj48L3BhdGg+CiAgICA8L2c+Cjwvc3ZnPgo=)](https://pytorch.org/)\n\nEnchanter is a library for machine learning tasks for comet.ml users.\n\n</div>\n\n---\n\n## Installation\nTo get started, [install PyTorch](https://pytorch.org) for your environment.\nThen install Enchanter in the following way:  \n\nTo install the stable release.\n```shell script\npip install enchanter\n```\n\nor\n\nTo install the latest(unstable) release. \n```shell script\npip install git+https://github.com/khirotaka/enchanter.git\n```\n\nIf you want to install with a specific branch, you can use the following.\n```shell script\n# e.g.) Install enchanter from develop branch.\npip install git+https://github.com/khirotaka/enchanter.git@develop\n```\n\n### Supported Platforms\nEnchanter supports:\n*   macOS 10.15\n*   Ubuntu 18.04\n\n## Getting Started\nTry your first Enchanter Program. To train a neural network written in PyTorch on Enchanter, use the `Runner`.  \nThere are 2 ways to define a `Runner`:\n\n1.  To use a `Runner` already implemented under `enchanter.tasks`\n2.  To define a custom `Runner` that inherit `enchanter.engine.BaseRunner`.\n\nLet\'s see how to use the `enchanter.tasks.ClassificationRunner`, which is the easiest way.\n\n### Training Neural Network\n\n```python\nimport comet_ml\nimport torch\nimport enchanter\n\nmodel = torch.nn.Linear(6, 10)\noptimizer = torch.optim.Adam(model.parameters())\n\nrunner = enchanter.tasks.ClassificationRunner(\n    model, \n    optimizer,\n    criterion=torch.nn.CrossEntropyLoss(),\n    experiment=comet_ml.Experiment()\n)\n\nrunner.add_loader("train", train_loader)\nrunner.train_config(epochs=10)\nrunner.run()\n```\n\nRegister a `torch.utils.data.DataLoader` with the `Runner` by using `.add_loader()`.  \nSet up the number of epochs using `.train_config()`, and execute `Runner` with `.run()`.\n\n### Training Unsupervised Time Series Feature Learning\nThe wonderful algorithms for unsupervised time series representation learning, adopted at [NeurIPS 2019](https://papers.nips.cc/paper/8713-unsupervised-scalable-representation-learning-for-multivariate-time-series), are now easily available.\n\nPlease prepare the following:\n\n1.  PyTorch Model that can output feature vectors of the same length regardless of the input series.\n2.  time series data consisting of `[N, F, L]`.\n3.  (Optional) A teacher label for each sample in `2.`\n\n\n```python\nimport comet_ml\nimport torch.nn as nn\nimport torch.optim as optim\nimport enchanter.tasks as tasks\nimport enchanter.addons.layers as L\n\n\nclass Encoder(nn.Module):\n    def __init__(self, in_features, mid_features, out_features):\n        super(Encoder, self).__init__()\n        self.conv = nn.Sequential(\n            L.CausalConv1d(in_features, mid_features, 3),\n            nn.LeakyReLU(),\n            L.CausalConv1d(mid_features, mid_features, 3),\n            nn.LeakyReLU(),\n            L.CausalConv1d(mid_features, mid_features, 3),\n            nn.LeakyReLU(),\n            nn.AdaptiveMaxPool1d(1)\n        )\n        self.fc = nn.Linear(mid_features, out_features)\n\n    def forward(self, x):\n        batch = x.shape[0]\n        out = self.conv(x).reshape(batch, -1)\n        return self.fc(out)\n\n\nexperiment = comet_ml.Experiment()\nmodel = Encoder(...)\noptimizer = optim.Adam(model.parameters())\n\nrunner = tasks.TimeSeriesUnsupervisedRunner(model, optimizer, experiment)\nrunner.add_loader("train", ...)\nrunner.run()\n```\n\nA teacher label is required for validation. Also, Use `enchanter.callbacks.EarlyStoppingForTSUS` for early stopping. \n \n### Hyper parameter searching using Comet.ml\n\n```python\nfrom comet_ml import Optimizer\n\nimport torch\nimport torch.nn as nn\nimport torch.optim as optim\nfrom sklearn.datasets import load_iris\n\nimport enchanter.tasks as tasks\nimport enchanter.addons as addons\nimport enchanter.addons.layers as layers\nfrom enchanter.utils import comet\n\n\nconfig = comet.TunerConfigGenerator(\n    algorithm="bayes",\n    metric="train_avg_loss",\n    objective="minimize",\n    seed=0,\n    trials=1,\n    max_combo=10\n)\n\nconfig.suggest_categorical("activation", ["addons.mish", "torch.relu", "torch.sigmoid"])\nopt = Optimizer(config.generate())\n\nx, y = load_iris(return_X_y=True)\nx = x.astype("float32")\ny = y.astype("int64")\n\n\nfor experiment in opt.get_experiments():\n    model = layers.MLP([4, 512, 128, 3], eval(experiment.get_parameter("activation")))\n    optimizer = optim.Adam(model.parameters())\n    runner = tasks.ClassificationRunner(\n        model, optimizer=optimizer, criterion=nn.CrossEntropyLoss(), experiment=experiment\n    )\n\n    runner.fit(x, y, epochs=1, batch_size=32)\n    runner.quite()\n\n    # or \n    # with runner:\n    #   runner.fit(...)\n    # or\n    #   runner.run()\n\n```\n\n### Training with Mixed Precision\nRunners with defined in `enchanter.tasks` are now support Auto Mixed Precision.  \nWrite the following.\n\n\n```python\nfrom torch.cuda import amp\nfrom enchanter.tasks import ClassificationRunner\n\n\nrunner = ClassificationRunner(...)\nrunner.scaler = amp.GradScaler()\n```\n\n\nIf you want to define a custom runner that supports mixed precision, do the following.\n```python\nfrom torch.cuda import amp\nimport torch.nn.functional as F\nfrom enchanter.engine import BaseRunner\n\n\nclass CustomRunner(BaseRunner):\n    # ...\n    def train_step(self, batch):\n        x, y = batch\n        with amp.autocast():        # REQUIRED\n            out = self.model(x)\n            loss = F.nll_loss(out, y)\n        \n        return {"loss": loss}\n\n\nrunner = CustomRunner(...)\nrunner.scaler = amp.GradScaler()\n```\n\nThat is, you can enable AMP by using `torch.cuda.amp.autocast()` in `.train_step()`, `.val_step()` and `.test_step()`.\n\n### with-statement training\n\n```python\nfrom comet_ml import Experiment\n\nimport torch\nimport torch.nn as nn\nimport torch.optim as optim\nfrom torch.utils.data import DataLoader\nfrom sklearn.datasets import load_iris\nfrom tqdm.auto import tqdm\n\nimport enchanter.tasks as tasks\nimport enchanter.engine.modules as modules\nimport enchanter.addons as addons\nimport enchanter.addons.layers as layers\n\n\nexperiment = Experiment()\nmodel = layers.MLP([4, 512, 128, 3], addons.mish)\noptimizer = optim.Adam(model.parameters())\n\nx, y = load_iris(return_X_y=True)\nx = x.astype("float32")\ny = y.astype("int64")\n\ntrain_ds = modules.get_dataset(x, y)\nval_ds = modules.get_dataset(x, y)\ntest_ds = modules.get_dataset(x, y)\n\ntrain_loader = DataLoader(train_ds, batch_size=32)\nval_loader = DataLoader(val_ds, batch_size=32)\ntest_loader = DataLoader(test_ds, batch_size=32)\n\nrunner = tasks.ClassificationRunner(model, optimizer, nn.CrossEntropyLoss(), experiment)\n\nwith runner:\n    for epoch in tqdm(range(10)):\n        with runner.experiment.train():\n            for train_batch in train_loader:\n                runner.optimizer.zero_grad()\n                train_out = runner.train_step(train_batch)\n                runner.backward(train_out["loss"])\n                runner.update_optimizer()\n    \n                with runner.experiment.validate(), torch.no_grad():\n                    for val_batch in val_loader:\n                        val_out = runner.val_step(val_batch)["loss"]\n                        runner.experiment.log_metric("val_loss", val_out)\n\n        with runner.experiment.test(), torch.no_grad():\n            for test_batch in test_loader:\n                test_out = runner.test_step(test_batch)["loss"]\n                runner.experiment.log_metric("test_loss", test_out)\n\n# The latest checkpoints (model_state & optim_state) are stored in comet.ml after the with statement.\n```\n\n## Graph visualization\n\n```python\nimport torch\nfrom enchanter.utils import visualize\nfrom enchanter.addons.layers import AutoEncoder\n\nx = torch.randn(1, 32)  # [N, in_features]\nmodel = AutoEncoder([32, 16, 8, 2])\nvisualize.with_netron(model, (x, ))\n```\n\n![netron_graph](docs/tutorial/assets/netron_viewer.png)\n\n## License\n[Apache License 2.0](LICENSE)\n',
    'author': 'Hirotaka Kawashima',
    'author_email': 'khirotaka@vivaldi.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://enchanter.readthedocs.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.5,<3.9',
}


setup(**setup_kwargs)



from .classification import ClassificationRunner
from .regression import RegressionRunner, AutoEncoderRunner
from .ensemble import HardEnsemble, SoftEnsemble, BaseEnsembleEstimator
from .ts_unsupervised import TimeSeriesUnsupervisedRunner

__all__ = [
    "ClassificationRunner",
    "RegressionRunner",
    "HardEnsemble",
    "SoftEnsemble",
    "BaseEnsembleEstimator",
    "TimeSeriesUnsupervisedRunner",
    "AutoEncoderRunner",
]



from typing import Union, Optional, Dict
from collections import OrderedDict
from time import ctime
from pathlib import Path
from copy import deepcopy

import torch
import torch.nn as nn


__all__ = ["RunnerIO"]


class RunnerIO:
    """
    A class responsible for loading and saving parameters such as PyTorch model weights and Optimizer state.

    """

    def __init__(self):
        self.model = NotImplemented
        self.optimizer = NotImplemented
        self.experiment = NotImplemented
        self.save_dir: Optional[str] = None

    def model_name(self) -> str:
        """
        fetch model name

        Returns: model name

        """
        if isinstance(self.model, nn.DataParallel) or isinstance(self.model, nn.parallel.DistributedDataParallel):
            model_name = self.model.module.__class__.__name__
        else:
            model_name = self.model.__class__.__name__

        return model_name

    def save_checkpoint(self) -> Dict[str, Union[Dict[str, torch.Tensor], dict]]:
        """
        A method to output model weights and Optimizer state as a dictionary.

        Returns:
            Returns a dictionary with the following keys and values.
                - ``model_state_dict``: model weights
                - ``optimizer_state_dict``: Optimizer state

        """
        if isinstance(self.model, nn.DataParallel) or isinstance(self.model, nn.parallel.DistributedDataParallel):
            model = self.model.module.state_dict()
        else:
            model = self.model.state_dict()

        checkpoint = {
            "model_state_dict": deepcopy(model),
            "optimizer_state_dict": deepcopy(self.optimizer.state_dict()),
        }
        return checkpoint

    def load_checkpoint(self, checkpoint: Dict[str, OrderedDict]):
        """
        Takes a dictionary with keys ``model_state_dict`` and ``optimizer_state_dict``
        and uses them to restore the state of the model and the Optimizer.

        Args:
            checkpoint:
                Takes a dictionary with the following keys and values.
                    - ``model_state_dict``: model weights
                    - ``optimizer_state_dict``: Optimizer state
        """
        self.model.load_state_dict(checkpoint["model_state_dict"])
        self.optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
        return self

    def save(self, directory: Optional[str] = None, epoch: Optional[int] = None, filename: Optional[str] = None):

        """
        Save the model and the Optimizer state file in the specified directory.

        Notes:
            ``enchanter_checkpoints_epoch_{}.pth`` file contains ``model_state_dict`` & ``optimizer_state_dict``.

        Args:
            directory (Optional[str]):
            epoch (Optional[int]):
            filename (Optional[str]):

        """
        if directory is None and self.save_dir:
            directory = self.save_dir

        if directory is None:
            if filename is None:
                raise ValueError("The argument `directory` or `filename` must be specified.")
            else:
                path = filename
        else:
            directory_path = Path(directory)
            if not directory_path.exists():
                directory_path.mkdir(parents=True)

            if epoch is None:
                epoch_str = ctime().replace(" ", "_")
            else:
                epoch_str = str(epoch)

            if not filename:
                filename = "enchanter_checkpoints_epoch_{}.pth".format(epoch_str)

            path = str(directory_path / filename)

        checkpoint = self.save_checkpoint()
        torch.save(checkpoint, path)

        model_name = self.model_name()
        self.experiment.log_model(model_name, str(path))

    def load(self, filename: str, map_location: str = "cpu"):
        """
        Restores the model and Optimizer state based on the specified file.

        Args:
            filename (str):
            map_location (str): default: 'cpu'

        """
        checkpoint = torch.load(filename, map_location=map_location)
        self.load_checkpoint(checkpoint)

        return self



from .activation import Mish, Swish, mish, FReLU1d, FReLU2d
from .optim_wrapper import TransformerOptimizer
from . import layers
from . import criterions


__all__ = ["Mish", "Swish", "FReLU1d", "FReLU2d", "mish", "TransformerOptimizer", "layers", "criterions"]



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from typing import Dict, Optional, List, Any, Union
from pprint import pformat
from pkg_resources import working_set


if "comet-ml" not in [d.project_name for d in working_set]:
    raise ImportError("You have to install `comet_ml` if you want to use `enchanter.utils.comet` module.")


__all__ = ["TunerConfigGenerator"]


class TunerConfigGenerator:
    def __init__(
        self,
        algorithm: str = "bayes",
        metric: str = "validate_avg_loss",
        objective: str = "minimize",
        seed: Optional[int] = None,
        max_combo: int = 0,
        grid_size: int = 10,
        min_sample_size: int = 100,
        retry_limit: int = 20,
        retry_assign_limit: int = 0,
        name: Optional[str] = None,
        trials: int = 1,
    ) -> None:
        """
        See https://www.comet.ml/docs/python-sdk/introduction-optimizer/ for a more detailed explanation of each argument.

        Args:
            algorithm: Specifies the algorithm used for parameter tuning. The supported algorithms are ``['grid','random','bayes']``.
            metric: Specify the value to be minimized / maximized. By default,``validate_avg_loss`` is specified.
            objective: Specifies whether to maximize / minimize metrics. Specify with ``['minimize','maximize']``.
            seed: Set the seed value. Not specified by default.
            max_combo: Integer. Limit on the combination of parameters to try (default is 0, meaning no limit)
            grid_size: Integer. Number of bins per parameter when creating a grid (default is 10).
            min_sample_size: Integer. Number of samples to help find a suitable grid range (default is 100).
            retry_limit: integer. A limit that attempts to create a unique set of parameters before suspending (default is 20).
            retry_assign_limit:
            name: String. A personalizable name associated with this search instance. (option)
            trials: Specifies the number of trials in a single experiment.
        """

        if algorithm not in ["grid", "bayes", "random"]:
            raise ValueError("The algorithms you can select are `random`, `bayes` and `grid`.")

        self.algorithm: str = algorithm
        self.spec: Dict[str, Any] = {
            "maxCombo": max_combo,
            "objective": objective,
            "metric": metric,
            "seed": seed,
            "gridSize": grid_size,
            "minSampleSize": min_sample_size,
            "retryLimit": retry_limit,
            "retryAssignLimit": retry_assign_limit,
        }
        self.name: Optional[str] = name
        self.trials: int = trials
        self.__params: Dict[str, Dict[str, Any]] = {}

    def suggest_categorical(self, name: str, values: List[str]):
        """
        A method for searching categorical variables.

        Args:
            name: Specify the name of the variable.
            values: I have a variable to search. Due to the specifications of comet.ml, it is necessary to give a list of strings.

        Examples:
            >>> import comet_ml
            >>> from enchanter.utils.comet import TunerConfigGenerator
            >>> config = TunerConfigGenerator()
            >>> config.suggest_categorical("activation", ["torch.relu", "torch.sigmoid", "torch.softmax"])
            >>> opt = comet_ml.Optimizer(config.generate())
            >>> for experiment in opt.get_experiments():
            >>>     activation = experiment.get_parameter(eval("activation"))

        """
        self.__params["{}".format(name)] = {
            "type": "categorical",
            "values": [str(value) for value in values],
        }
        return self

    def __suggest(
        self,
        name: str,
        min_value: Union[float, int],
        max_value: Union[float, int],
        dtype: Optional[type],
        scaling: str,
        **kwargs,
    ) -> None:
        if dtype is None:
            if type(min_value) is type(max_value):
                dtype = type(min_value)
            else:
                raise Exception("`min_value` and `max_value` must be same data type.")

        if dtype is float:
            dtype_str = "float"
        elif dtype is int:
            dtype_str = "integer"
        else:
            raise Exception("`dtype` must be specified as `int` or `float` .")

        self.__params["{}".format(name)] = {
            "type": dtype_str,
            "min": min_value,
            "max": max_value,
            "scalingType": scaling,
        }

        for key in ["mu", "sigma"]:
            if key in kwargs:
                self.__params["{}".format(name)][key] = kwargs[key]

    def suggest_linear(
        self,
        name: str,
        min_value: Union[float, int],
        max_value: Union[float, int],
        dtype: Optional[type] = None,
    ):
        """
        If Integer, independent distribution. else if float the same as uniform.

        Args:
            name:
            min_value:
            max_value:
            dtype:

        Returns:

        """
        self.__suggest(name, min_value, max_value, dtype, "linear")
        return self

    def suggest_uniform(
        self,
        name: str,
        min_value: Union[float, int],
        max_value: Union[float, int],
        dtype: Optional[type] = None,
    ):
        """
        This is a method for sampling and searching variables from a uniform distribution.

        Args:
            name: Variable name
            min_value: minimum value
            max_value: Maximum value
            dtype: Data type. If not specified, it is automatically estimated from the values of ``min_value`` and` `max_value``.

        Examples:
            >>> import comet_ml
            >>> config = TunerConfigGenerator()
            >>> config.suggest_uniform("uniform", 0.0, 1.0)
            >>> opt = comet_ml.Optimizer(config.generate())
            >>> for experiment in opt.get_experiments():
            >>>     uniform = experiment.get_parameter("uniform")

        """
        self.__suggest(name, min_value, max_value, dtype, "uniform")
        return self

    def suggest_normal(
        self,
        name: str,
        min_value: Union[float, int],
        max_value: Union[float, int],
        mu: float = 0.0,
        sigma: float = 1.0,
        dtype: Optional[type] = None,
    ):
        """
        This is a method for sampling and searching variables from a normal distribution.

        Args:
            name: Variable name
            min_value: minimum value
            max_value: Maximum value
            mu: Normal distribution μ
            sigma: Normal distribution σ
            dtype: Data type. If not specified, it is automatically estimated from the values of ``min_value`` and` `max_value``.

        Examples:
            >>> import comet_ml
            >>> config = TunerConfigGenerator()
            >>> config.suggest_normal("normal", 0.0, 1.0)
            >>> opt = comet_ml.Optimizer(config.generate())
            >>> for experiment in opt.get_experiments():
            >>>     normal = experiment.get_parameter("normal")

        """
        self.__suggest(name, min_value, max_value, dtype, "normal", mu=mu, sigma=sigma)
        return self

    def suggest_lognormal(
        self,
        name: str,
        min_value: Union[float, int],
        max_value: Union[float, int],
        mu: float = 0.0,
        sigma: float = 1.0,
        dtype: Optional[type] = None,
    ):
        """
        A method for sampling and searching variables from a lognormal distribution.

        Args:
            name: Variable name
            min_value: minimum value
            max_value: Maximum value
            mu: Normal distribution μ
            sigma: Normal distribution σ
            dtype: Data type. If not specified, it is automatically estimated from the values of ``min_value`` and` `max_value``.

        Examples:
            >>> import comet_ml
            >>> config = TunerConfigGenerator()
            >>> config.suggest_lognormal("lognormal", 0.0, 1.0)
            >>> opt = comet_ml.Optimizer(config.generate())
            >>> for experiment in opt.get_experiments():
            >>>     lognormal = experiment.get_parameter("lognormal")

        """
        self.__suggest(name, min_value, max_value, dtype, "lognormal", mu=mu, sigma=sigma)
        return self

    def suggest_loguniform(
        self,
        name: str,
        min_value: Union[float, int],
        max_value: Union[float, int],
        dtype: Optional[type] = None,
    ):
        """
        This is a method for sampling and searching variables from a uniform distribution.

        Args:
            name: Variable name
            min_value: minimum value
            max_value: Maximum value
            dtype: Data type. If not specified, it is automatically estimated from the values of ``min_value`` and` `max_value``.

        Examples:
            >>> import comet_ml
            >>> config = TunerConfigGenerator()
            >>> config.suggest_loguniform("loguniform", 0.0, 1.0)
            >>> opt = comet_ml.Optimizer(config.generate())
            >>> for experiment in opt.get_experiments():
            >>>     loguniform = experiment.get_parameter("loguniform")

        """
        self.__suggest(name, min_value, max_value, dtype, "loguniform")
        return self

    def suggest_discrete(self, name: str, values: List[Union[float, int]]):
        """
        This method is used to search for the specified numeric type variable.

        Args:
            name: Variable name
            values: A list of numeric elements

        Examples:
            >>> import comet_ml
            >>> config = TunerConfigGenerator()
            >>> config.suggest_discrete("discrete", [10, 20, 30])
            >>> opt = comet_ml.Optimizer(config.generate())
            >>> for experiment in opt.get_experiments():
            >>>     discrete = experiment.get_parameter("discrete")

        """
        self.__params["{}".format(name)] = {"type": "discrete", "values": values}
        return self

    def generate(self) -> Dict[str, Any]:
        """
        A method for generating a Config for ``comet_ml.Optimizer``.

        Examples:
            >>> cfg = TunerConfigGenerator()
            >>> cfg.suggest_discrete("discrete", [10, 20, 30])
            >>> cfg_dict = cfg.generate()
        """

        config = {
            "algorithm": self.algorithm,
            "parameters": self.__params,
            "spec": self.spec,
            "trials": self.trials,
            "name": self.name,
        }
        return config

    def export(self, filename: str) -> None:
        """
        It is a method to save the created Config as a file.

        """
        with open(filename, "w") as f:
            f.write(self.__repr__())

    def __repr__(self):
        return pformat(self.generate(), indent=1)
